/**********************************************************************
**
** Copyright (C) 2003 Germain Garand <germain@ebooksfrance.org>
**
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
**********************************************************************/

#include "perlaction.h"
#include <qregexp.h>
#include <qfile.h>
#include <qlineedit.h>
#include <qlistbox.h>
#include <qmessagebox.h>
#include <qradiobutton.h>
#include <time.h>
#include <qapplication.h>
#include "perlmainprojectfile.h"
#include "editorinterfaceimpl.h"

QProcess* PerlAction::perl_process = 0;

PerlAction::PerlAction( QObject* p, DesignerInterface* d, int t ): QAction( p ), dIface( d )
{
    if( dIface )
	dIface->addRef();
    if( t == RunForm )
    {
	connect( this, SIGNAL( activated() ), this, SLOT( runForm() ) );
	setText( "Run current form");
	setIconSet( QIconSet( QPixmap::fromMimeSource( "perlqt.png" ) ) );
    }
    else
    {
        connect( this, SIGNAL( activated() ), this, SLOT( runProject() ) );
	setText( "Build and run current project" );
	setIconSet( QIconSet( QPixmap::fromMimeSource( "perlqtblue2.png" ) ) );
    }

}

PerlAction::~PerlAction()
{

}

bool PerlAction::checkProcess()
{
    if( perl_process && perl_process->isRunning() )
    {
	int u = QMessageBox::information( dynamic_cast<QWidget *>(this->parent()), tr("External Perl Process"),
		tr( "<qt>A Perl process is already running...<br> Terminate?</qt>"),
		QMessageBox::Yes, QMessageBox::Cancel );
	switch( u )
	{
	    case QMessageBox::Yes:
		perl_process->tryTerminate();
		perl_process->deleteLater();
		perl_process = 0;
		break;
	    default:
	        return false;
	}
    }
    else if( perl_process )
    {
	perl_process->deleteLater();
	perl_process = 0;
    }
    return true;
}

void PerlAction::runForm()
{
    if( !dIface || !dIface->currentForm() || !( dIface->currentProject() &&
	dIface->currentProject()->language() == "Perl" ) || !checkProcess() )
	return;

    QString fname = dIface->currentForm()->fileName();
    if( dIface->currentForm()->isModified() || fname.isNull() || ! QFile( fname ).exists() )
    {
	QMessageBox::information( dynamic_cast<QWidget *>(this->parent()), tr("Run Form") ,
			tr( "Please save your work first" ) );
	return;
    }
    QStringList lst;
    QString puicpath = Config::puicPath( "/Trolltech/PerlEditor" );
    QFileInfo file( fname );
    QString path = file.absFilePath();
    path = path.left( path.length()  - file.fileName().length() );
    lst << puicpath << "-x" << file.fileName();
    QProcess puic( lst );
    puic.setWorkingDirectory( QDir( path ) );
    bool success = puic.start();
    if( !success )
    {
	QMessageBox::information( dynamic_cast<QWidget *>(this->parent()), tr("Run Form") ,
		 tr( "<qt>Couldn't run puic.<br>"
		     "Please check external applications pathes in global preferences</qt>" ) );
	return;
    }

    struct timespec ts;
    ts.tv_sec = 0;
    ts.tv_nsec = 40000;
    while( puic.isRunning() )
    {
	nanosleep( &ts, NULL );
	qApp->processEvents();
    }

    QString code;
    while( puic.isRunning() || puic.canReadLineStdout() )
        if( puic.canReadLineStdout() )
	    code += puic.readLineStdout() + "\n";
    if( code.isNull() )
        return;
    QString perlpath = Config::perlPath( "/Trolltech/PerlEditor" );
    QStringList lst2( QStringList::split( " ", perlpath) );
    lst2 << "-w" << "-e" << "BEGIN { select(STDOUT); $| = 1 }\n" + code ;
    perl_process = new QProcess( lst2 );
    connect( perl_process, SIGNAL( readyReadStdout() ), this, SLOT( outputStdout() ) );
    connect( perl_process, SIGNAL( readyReadStderr() ), this, SLOT( outputStderr() ) );
    connect( perl_process, SIGNAL( processExited() ), this, SLOT( processTerminated() ) );
    if ( perl_process && !perl_process->start() )
    {
	perl_process->deleteLater();
	perl_process = 0;
    }

}

void PerlAction::runProject()
{
    if( !dIface || !( dIface->currentProject() && dIface->currentProject()->language() == "Perl" ) ||
	!checkProcess() )
        return;
//    qDebug( "projname: %s", dIface->currentProject()->fileName().latin1() );
//    qDebug( "main-form: %s",dIface->currentProject()->customSetting("MAIN_FORM").latin1() );
    DesignerProject *pro= dIface->currentProject();
    if( pro->fileName().isEmpty() || pro->fileName() == ".pro" )
    {
	QMessageBox::information( dynamic_cast<QWidget *>(this->parent()), tr("Run Project") ,
	    tr( "Please select or create a project" ) );
	return;
    }
    QFileInfo projfile( pro->fileName() );
    if( ! projfile.exists() )
    {
	QMessageBox::information( dynamic_cast<QWidget *>(this->parent()), tr("Run Project") ,
	    tr( "Can't find project file '%1' in current directory" ).arg( pro->fileName() ) );
	return;
    }

    QPtrList<DesignerFormWindow> fl = pro->formList();
    QPtrListIterator<DesignerFormWindow> it( fl );
    DesignerFormWindow *dfw;
    while ( (dfw = it.current()) != 0 )
    {
        ++it;
//	qDebug("filename: %s", dfw->fileName().latin1() );
        if( dfw->isModified() ||
	    dfw->fileName().isNull() || ! QFile( dfw->fileName() ).exists() )
	{
	    QMessageBox::information( dynamic_cast<QWidget *>(this->parent()), tr("Run Project") ,
			tr( "Some files are modified. Please save them first" ) );
	    return;
	}
    }
    bool mod = false;
    QValueList< QGuardedPtr<PerlEditor> > edlist = EditorInterfaceImpl::editorList();
    for( QValueListIterator< QGuardedPtr<PerlEditor> > it = edlist.begin(); it != edlist.end(); )
    {
        PerlEditor *ed = *it;
	if( ed && ed->isModified() )
	{
	    mod = true;
	    break;
	}
	++it;
    }
    if( mod )
    {
	QMessageBox::information( dynamic_cast<QWidget *>(this->parent()), tr("Run Project") ,
	tr( "Some files are modified. Please save them first" ) );
	return;
    }

    QString mainfile, mainform;
    bool generateMain  = false;

    QString projpath = projfile.absFilePath();
    projpath = projpath.left( projpath.length() - projfile.fileName().length() );
    if( pro->customSetting("PROGRAM_NAME").isEmpty() ||
      (!pro->customSetting("MAIN_FORM").isEmpty() &&
        pro->formFileName( pro->customSetting("MAIN_FORM") ).isEmpty() ) )
    {
        PerlMainProjectFile dia( 0, 0, TRUE );
        dia.setup( dIface );
        if ( dia.exec() == QDialog::Accepted )
	{
	    mainfile = dia.editFileName->text();
	    if( !mainfile.isEmpty() )
	        pro->setCustomSetting( "PROGRAM_NAME", mainfile );
	    if( dia.radioGenerate->isChecked() )
	    {
	        generateMain = true;
	        mainform = dia.listForms->text( dia.listForms->currentItem() );
		pro->setCustomSetting( "MAIN_FORM", mainform );
	    }
	    else
		pro->setCustomSetting( "MAIN_FORM", QString::null );
        }
        else
            return;
    }
    else
    {
	mainfile = pro->customSetting("PROGRAM_NAME");
	if( !pro->customSetting("MAIN_FORM").isEmpty() )
	{
	    generateMain = true;
	    mainform = pro->customSetting("MAIN_FORM");
	}
    }

    QFile mf( projpath + mainfile );

    if( !generateMain )
    {
    	if( mainfile.isEmpty() || !mf.exists() )
	{
	    QMessageBox::information( dynamic_cast<QWidget *>(this->parent()), tr("Error") ,
			 tr( "<qt>There is no '%1' file in current project's <br> main directory '%2'...</qt>" )
			 .arg( mainfile ).arg( projpath ) );
	    return;
	}
    }
    else if( mf.exists() && mf.open( IO_ReadOnly) )
    {
        QTextStream stream( &mf );
	QString s = stream.readLine();
	mf.close();
	if( ! s.startsWith( "# Form implementation" ) )
	{
	     int u = QMessageBox::information( dynamic_cast<QWidget *>(this->parent()),
			tr("Run Project"),
			tr( "<qt>File %1 exists in %2, <br>and has not been auto-generated... Overwrite?</qt>")
			.arg( mainfile ).arg( projpath ),
			QMessageBox::Yes, QMessageBox::Cancel );
	    switch( u )
	    {
	        case QMessageBox::Yes:
		    break;
		default:
	            return;
	    }
	}
    }

    QStringList formnames = pro->formNames();
    QStringList fnames;
    QString puicpath = Config::puicPath( "/Trolltech/PerlEditor" );
    if( puicpath.simplifyWhiteSpace().isEmpty() )
    {
	QMessageBox::information( dynamic_cast<QWidget *>(this->parent()), tr("Run Project") ,
		 tr( "<qt>Couldn't run puic.<br>"
		     "Please check external applications pathes in global preferences</qt>" ) );
	return;
    }
    for( QStringList::Iterator it= formnames.begin(); it != formnames.end() ; ++it )
    {
        QString fname = pro->formFileName( *it );
//	qDebug( "form: %s fname: %s", (*it).latin1(), fname.latin1() );
        fnames << fname;
        QStringList lst( QStringList::split( " ", puicpath) );
        QFileInfo file( fname );
        QString path = file.absFilePath();
        path = path.left( path.length()  - file.fileName().length() );
	if( generateMain && mainform == *it )
	{
	    lst << "-x" << "-o" << projpath + mainfile << file.fileName();
	}
	else
            lst << "-o" << (*it) + ".pm" << file.fileName();
//        qDebug( "puic: %s", lst.join( " " ).latin1() );
	QProcess puic( lst );
        puic.setWorkingDirectory( QDir( (generateMain && mainfile == *it) ? projpath : path ) );
        bool success = puic.start();
        if( !success )
        {
	    QMessageBox::information( dynamic_cast<QWidget *>(this->parent()), tr("Run Project") ,
			 tr( "An error occured while running puic on file %1" ).arg( fname ) );
	    return;
        }

	struct timespec ts;
	ts.tv_sec = 0;
	ts.tv_nsec = 40000;
	while( puic.isRunning() )
	{
		nanosleep( &ts, NULL );
		qApp->processEvents();
	}
    }

    QString perlpath = Config::perlPath( "/Trolltech/PerlEditor" );
    QStringList lst2( QStringList::split( " ", perlpath) );
    QString code = "BEGIN { select(STDOUT); $| = 1 }\n# line 1 \"" + mainfile + "\"\n";
    {
        QFile main( projpath + mainfile );
        if( !main.exists() || !main.open( IO_ReadOnly ) )
        {
	    QMessageBox::information( dynamic_cast<QWidget *>(this->parent()), tr("Run Project") ,
			 tr( "Couldn't open generated Main-File %1 in %2" ).arg( mainfile ).arg( projpath ) );
	    return;
        }
        QTextStream stream( &main );
        while( !stream.eof() )
	    code += stream.readLine() + "\n";
	main.close();
    }
    lst2 << "-w" << "-e" << code ;
    perl_process = new QProcess( lst2 );
    perl_process->setWorkingDirectory( projpath );
    connect( perl_process, SIGNAL( readyReadStdout() ), this, SLOT( outputStdout() ) );
    connect( perl_process, SIGNAL( readyReadStderr() ), this, SLOT( outputStderr() ) );
    connect( perl_process, SIGNAL( processExited() ), this, SLOT( processTerminated() ) );

    if ( perl_process && !perl_process->start() )
    {
    	qDebug( "failed to run perl process." );
	perl_process->deleteLater();
	perl_process = 0;
    }
}

void PerlAction::outputStdout()
{
    if ( !dIface || !perl_process )
	return;
    DesignerOutputDock* dock = dIface->outputDock();
    if( !dock )
        return;
    while( perl_process->canReadLineStdout() )
    {
	QString out = perl_process->readLineStdout();
	out.replace( "<", "&lt;" );
	out.replace( ">", "&gt;" );
	dock->appendDebug( "<font color=\"blue\"><b>STDOUT&gt;</b></font> " + out );
    }
}

void PerlAction::outputStderr()
{
    if ( !dIface || !perl_process )
	return;
    DesignerOutputDock* dock = dIface->outputDock();
    if( !dock )
        return;
//    QRegExp error( "at \\S+ line (\\d+)\\." );
    while( perl_process->canReadLineStderr() )
    {
    	QString err = perl_process->readLineStderr();
//	if( err.find( error ) != -1 )
//	    dock->appendError( err, error.cap(1).toInt() ); //- we can't do that... this func is buggy ;(
	err.replace( "<", "&lt;" );
	err.replace( ">", "&gt;" );
	dock->appendDebug( "<font color=\"red\"><b>STDERR&gt;</b> " + err + "</font>");
    }
}

void PerlAction::processTerminated()
{
    if ( !dIface || !perl_process )
	return;
    DesignerOutputDock* dock = dIface->outputDock();
    if( !dock )
        return;
    QString exit;
    if( !perl_process->normalExit() )
    	exit = tr("the perl process exited abnormally.");
    else if( !perl_process->exitStatus() )
	exit = tr("the perl process exited successfully");
    else
        exit = tr("the perl process exited with a non-zero status: %1")
		.arg( QString::number( perl_process->exitStatus() ) );
    dIface->showStatusMessage( exit, 2000 );
    dock->appendDebug( exit.prepend("<b>").append("</b>") );
}
