/**********************************************************************
**
** Copyright (C) 2000 Trolltech AS.  All rights reserved.
** Copyright (C) 2003 Germain Garand <germain@ebooksfrance.org>
**
** This file is part of Qt Designer.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/

#include "syntaxhighliter_perl.h"
#include "paragdata.h"
#include "qstring.h"
#include "qstringlist.h"
#include "qcstring.h"
#include "qmap.h"
#include "qapplication.h"
#include "qregexp.h"
#include "qmemarray.h"

const char * const SyntaxHighlighter_Perl::keywords[] = {
    // Perl keywords
    "if",
    "unless",
    "else",
    "elsif",
    "while",
    "until",
    "for",
    "each",
    "foreach",
    "next",
    "last",
    "break",
    "continue",
    "return",
    "use",
    "no",
    "require",
    "my",
    "our",
    "local",
    "BEGIN",
    "END",
    "package",
    "sub",
    "do",
    "__END__",
    "__DATA__",
    "__FILE__",
    "__LINE__",
    "__PACKAGE__",
    // additional "keywords" intoduced by PerlQt
    "CAST",
    "SIGNAL",
    "SLOT",
    "DCOP",
    "emit",
    // end of array
    0
};


const char * const SyntaxHighlighter_Perl::functions[] = {
    "abs",
    "accept",
    "alarm",
    "atan2",
    "bind",
    "binmode",
    "bless",
    "caller",
    "chdir",
    "chmod",
    "chomp",
    "chop",
    "chown",
    "chr",
    "chroot",
    "close",
    "closedir",
    "connect",
    "cos",
    "crypt",
    "dbmclose",
    "dbmopen",
    "defined",
    "delete",
    "die",
    "dump",
    "endgrent",
    "endhostent",
    "endnetent",
    "endprotoent",
    "endpwent",
    "endservent",
    "eof",
    "eval",
    "exec",
    "exists",
    "exit",
    "exp",
    "fcntl",
    "fileno",
    "flock",
    "fork",
    "format",
    "formline",
    "getc",
    "getgrent",
    "getgrgid",
    "getgrnam",
    "gethostbyaddr",
    "gethostbyname",
    "gethostent",
    "getlogin",
    "getnetbyaddr",
    "getnetbyname",
    "getnetent",
    "getpeername",
    "getpgrp",
    "getppid",
    "getpriority",
    "getprotobyname",
    "getprotobynumber",
    "getprotoent",
    "getpwent",
    "getpwnam",
    "getpwuid",
    "getservbyname",
    "getservbyport",
    "getservent",
    "getsockname",
    "getsockopt",
    "glob",
    "gmtime",
    "goto",
    "grep",
    "hex",
    "import",
    "index",
    "int",
    "ioctl",
    "join",
    "keys",
    "kill",
    "last",
    "lc",
    "lcfirst",
    "length",
    "link",
    "listen",
    "localtime",
    "lock",
    "log",
    "lstat",
    "map",
    "mkdir",
    "msgctl",
    "msgget",
    "msgrcv",
    "msgsnd",
    "oct",
    "open",
    "opendir",
    "ord",
    "pack",
    "package",
    "pipe",
    "pop",
    "pos",
    "print",
    "printf",
    "prototype",
    "push",
    "quotemeta",
    "rand",
    "read",
    "readdir",
    "readline",
    "readlink",
    "recv",
    "redo",
    "ref",
    "rename",
    "reset",
    "return",
    "reverse",
    "rewinddir",
    "rindex",
    "rmdir",
    "scalar",
    "seek",
    "seekdir",
    "select",
    "semctl",
    "semget",
    "semop",
    "send",
    "setgrent",
    "sethostent",
    "setnetent",
    "setpgrp",
    "setpriority",
    "setprotoent",
    "setpwent",
    "setservent",
    "setsockopt",
    "shift",
    "shmctl",
    "shmget",
    "shmread",
    "shmwrite",
    "shutdown",
    "sin",
    "sleep",
    "socket",
    "socketpair",
    "sort",
    "splice",
    "split",
    "sprintf",
    "sqrt",
    "srand",
    "stat",
    "study",
    "sub",
    "substr",
    "symlink",
    "syscall",
    "sysread",
    "sysseek",
    "system",
    "syswrite",
    "tell",
    "telldir",
    "tie",
    "time",
    "times",
    "truncate",
    "uc",
    "ucfirst",
    "umask",
    "undef",
    "unlink",
    "unpack",
    "unshift",
    "untie",
    "utime",
    "values",
    "vec",
    "wait",
    "waitpid",
    "wantarray",
    "warn",
    "write",
    // end of array
    0
};

static QMap<int, QMap<QString, int > > *wordMap = 0;
//static QMap<int, QMap<QString, int > > *funcMap = 0;

static const char *alpha = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
static const char *num = "0123456789";

static inline int balancedCharFor( int c )
{
    switch( c )
    {
        case '{': return '}';
	case '(': return ')';
	case '[': return ']';
	default: return c;
    }
}

SyntaxHighlighter_Perl::SyntaxHighlighter_Perl()
    : QTextPreProcessor(), lastFormat( 0 ), lastFormatId( -1 )
{
    regExpMatcher = new QRegExp( "[ms]|tr|q[qwxr]?" );
    slashREMatcher = new QRegExp( "=~\\s*/" );

    QFont f( qApp->font() );

    addFormat( Standard, new QTextFormat( f, Qt::black ) );
    addFormat( Number, new QTextFormat( f, Qt::blue ) );
    addFormat( String, new QTextFormat( f, Qt::red ) );
    addFormat( Type, new QTextFormat( f, Qt::darkMagenta ) );
    f.setBold( true );
    addFormat( Keyword, new QTextFormat( f, Qt::black ) );
    f.setBold( false );
    addFormat( CoreFunction, new QTextFormat( f, Qt::darkBlue ) );
    addFormat( Hash, new QTextFormat( f, Qt::red ) );
    addFormat( Label, new QTextFormat( f, Qt::magenta ) );
    addFormat( Scalar, new QTextFormat( f, Qt::darkRed ) );
    addFormat( RegExp, new QTextFormat( f, Qt::darkGreen ) );
    f.setFamily( "times" );
    addFormat( Comment, new QTextFormat( f, Qt::gray ) );

    if ( wordMap )
	return;

    wordMap = new QMap<int, QMap<QString, int> >;
    int len;
    for ( int i = 0; keywords[ i ]; ++i ) {
	len = strlen( keywords[ i ] );
	if ( !wordMap->contains( len ) )
	    wordMap->insert( len, QMap<QString, int >() );
	QMap<QString, int> &map = wordMap->operator[]( len );
	map[ keywords[ i ] ] = Keyword;
    }
    for ( int i = 0; functions[ i ]; ++i ) {
	len = strlen( functions[ i ] );
	if ( !wordMap->contains( len ) )
	    wordMap->insert( len, QMap<QString, int >() );
	QMap<QString, int> &map = wordMap->operator[]( len );
	map[ functions[ i ] ] = CoreFunction;
    }
}

SyntaxHighlighter_Perl::~SyntaxHighlighter_Perl()
{
    delete regExpMatcher;
}

static int string2Id( const QString &s )
{
    if ( s == "Standard" )
	return SyntaxHighlighter_Perl::Standard;
    if ( s == "Comment" )
	return SyntaxHighlighter_Perl::Comment;
    if ( s == "Number" )
	return SyntaxHighlighter_Perl::Number;
    if ( s == "String" )
	return SyntaxHighlighter_Perl::String;
    if ( s == "Qt Namespace" )
	return SyntaxHighlighter_Perl::Type;
    if ( s == "Core Function" )
	return SyntaxHighlighter_Perl::CoreFunction;
    if ( s == "Hash Key" )
	return SyntaxHighlighter_Perl::Hash;
    if ( s == "Label" )
	return SyntaxHighlighter_Perl::Label;
    if ( s == "Keyword" )
	return SyntaxHighlighter_Perl::Keyword;
    if ( s == "Scalar" )
	return SyntaxHighlighter_Perl::Scalar;
    if ( s == "Regular Expression" )
	return SyntaxHighlighter_Perl::RegExp;
    return SyntaxHighlighter_Perl::Standard;
}

void SyntaxHighlighter_Perl::updateStyles( const QMap<QString, ConfigStyle> &styles )
{
    for ( QMap<QString, ConfigStyle>::ConstIterator it = styles.begin(); it != styles.end(); ++it ) {
	int id = string2Id( it.key() );
	QTextFormat *f = format( id );
	if ( !f )
	    continue;
	f->setFont( (*it).font );
	f->setColor( (*it).color );
    }
}

static inline int checkSimpleRegExp( QTextParagraph *string, int i )
{
    int j = i;
    QByteArray n;
    n.setRawData( num, 10 );
    // first, check backward for a regexp operator, or the beginning of a block/parenthese
    for(; j && (string->at( j-1 )->c==' ' || string->at( j-1 )->c=='\t') ; j-- );
    if( !j ) {
	// mmh nothing, let's see if the other side looks like a denominator
	for( j=i+1 ; j < (string->length()-1) && (string->at( j )->c==' ' || string->at( j )->c=='\t') ; j++);
	if( string->at( j )->c == '\\' || string->at( j )->c == '[' || string->at( j )->c == '/' ||
	    (string->at( j )->c != '$' && string->at( j )->c != '(' && n.bsearch( string->at( j )->c ) == -1) )
	    j = 1;
	else j = 0;
    } else if( ((j-1)>0 && (string->at( j-1 )->c == '~' && string->at( j-2 )->c == '=' )) ||
	string->at( j-1 )->c == '{' || string->at( j-1 )->c == '(' || string->at( j-1 )->c==';' )
	j = 1;
    else j = 0;
    n.resetRawData( num, 10 );
    return j;
}

void SyntaxHighlighter_Perl::process( QTextDocument *doc, QTextParagraph *string, int, bool invalidate )
{

    QTextFormat *formatStandard = format( Standard );
    QTextFormat *formatComment = format( Comment );
    QTextFormat *formatNumber = format( Number );
    QTextFormat *formatString = format( String );
    QTextFormat *formatType = format( Type );
    QTextFormat *formatHash = format( Hash );
    QTextFormat *formatLabel = format( Label );
    QTextFormat *formatScalar = format( Scalar );
    QTextFormat *formatRegExp = format( RegExp );

    // states
    const int StateStandard = 0;
    const int StateStartScalar = 1;
    const int StateScalar = 2;
    const int StateProtScalar = 3;
    const int StateString = 4;
    const int StateString2 = 5;
    const int StateNumber = 6;
    const int StateComment = 7;
    const int StateHash = 8;
    const int StateMayBeRegExp = 9;
    const int StateRegExp = 10;

    // flags
    const int Pop = 0x100;
    const int PopOnce = 0x101;
    const int PopTwice = 0x102;
    const int PopThree = 0x103;
    const int After = 0x200;
    const int Push = 0x400;
    //    const int Again = 0x800;

    const int PopMask = 0xFF;

    // tokens
    const int InputAlpha = 0;
    const int InputNumber = 1;
    const int InputDollar = 2;
    const int InputSlash = 3;
    const int InputParen = 4;
    const int InputSpace = 5;
    const int InputHash = 6;
    const int InputQuotation = 7;
    const int InputApostrophe = 8;
    const int InputSep = 9;
    const int InputPercent = 10;
    const int InputArobas = 11;
    const int InputBraceLeft = 12;
    const int InputBraceRight = 13;

    static unsigned short table[ 11 ][ 14 ] = {
    //    Alpha               Number           Dollar               Slash                  Parenthesis      Space            Hash               Double-quote      Apostrophe         Separator         Percent           Arobas             BraceLeft               BraceRight
	{ StateStandard,      StateNumber,     StateStartScalar,    StateStandard,         StateStandard,   StateStandard,   StateComment,      StateString,      StateString2,      StateStandard,    StateStartScalar, StateStartScalar,  StateStandard | Push,   PopOnce | After  }, // StateStandard
	{ StateScalar,        StateScalar,     StateScalar,         StateScalar,           StateScalar,     PopOnce,         StateScalar,       StateScalar,      StateScalar,       StateScalar,      StateScalar,      StateScalar,       StateProtScalar,        StateScalar      }, // StateStartScalar
	{ StateScalar,        StateScalar,     StateScalar,         PopTwice,              PopTwice,        PopTwice,        PopTwice,          PopTwice,         PopTwice,          PopTwice,         PopTwice,         PopTwice,          StateHash,              PopTwice         }, // StateScalar
	{ StateProtScalar,    StateProtScalar, StateStartScalar,    StateProtScalar,       StateProtScalar, StateProtScalar, StateProtScalar,   StateString,      StateProtScalar,   StateProtScalar,  StateStartScalar, StateStartScalar,  StateStandard,          PopTwice | After }, // StateProtScalar
	{ StateString,        StateString,     StateStartScalar,    StateString,           StateString,     StateString,     StateString,       PopOnce | After,  StateString,       StateString,      StateStartScalar, StateStartScalar,  StateString,            StateString      }, // StateString
	{ StateString2,       StateString2,    StateString2,        StateString2,          StateString2,    StateString2,    StateString2,      StateString2,     PopOnce | After,   StateString2,     StateString2,     StateString2,      StateString2,           StateString2     }, // StateString2
	{ StateNumber,        StateNumber,     PopOnce,             PopOnce,               PopOnce,         PopOnce,         StateComment,      StateString,      PopOnce,           PopOnce,          PopOnce,          PopOnce,           PopOnce,                PopOnce          }, // StateNumber
	{ StateComment,       StateComment,    StateComment,        StateComment,          StateComment,    StateComment,    StateComment,      StateComment,     StateComment,      StateComment,     StateComment,     StateComment,      StateComment,           StateComment     }, // StateComment
	{ StateHash,          StateHash,       StateStartScalar,    StateHash,             StateHash,       StateHash,       StateHash,         StateString,      StateString2,      StateHash,        StateStartScalar, StateStartScalar,  StateStandard,          PopThree | After }, // StateHash
	{ PopOnce,            PopOnce,         StateMayBeRegExp,    StateMayBeRegExp,      StateMayBeRegExp,PopOnce,         StateMayBeRegExp,  StateMayBeRegExp, StateMayBeRegExp,  StateMayBeRegExp, StateMayBeRegExp, StateMayBeRegExp,  StateMayBeRegExp,       PopOnce          }, // StateMayBeRegExp
	{ StateRegExp,        StateRegExp,     StateRegExp,         StateRegExp,           StateRegExp,     StateRegExp,     StateRegExp,       StateRegExp,      StateRegExp,       StateRegExp,      StateRegExp,      StateRegExp,       StateRegExp,            StateRegExp,     }  // StateRegExp
    };

    QString buffer;
    static QMemArray<int> stack( 16 );
    static int sp;

    int state = stack[0] = StateStandard;
    sp = 0;
    if ( string->prev() ) {
	if ( string->prev()->endState() == -1 )
	    process( doc, string->prev(), 0, FALSE );
	state = string->prev()->endState();
    }
    int input;
    int i = 0;
    bool lastWasBackSlash = FALSE;
    bool makeLastStandard = FALSE;

    ParagData *paragData = (ParagData*)string->extraData();
    if ( paragData )
	paragData->parenList.clear();
    else
	paragData = new ParagData;
    if ( string->prev() && string->prev()->extraData() )
    {
	ParagData *prevParagData = (ParagData*)string->prev()->extraData();
	//if( prevParagData->endChar )
	    paragData->endChar = prevParagData->endChar;
    }

    //QString alphabeth = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
    QString mathChars = "xXeE";
    //QString numbers = "0123456789";
    QByteArray alphabet;
    QByteArray numbers;
    alphabet.setRawData( alpha, 52 );
    numbers.setRawData( num, 10 );

    bool questionMark = FALSE;
    QChar lastChar;
    for (;;) {
	QChar c = string->at( i )->c;

	if ( lastWasBackSlash && state != StateStandard ) {
	    input = InputSep;
	} else {
	    switch ( c ) {
	    case '$':
		input = InputDollar;
		break;
	    case '{':
	        input = InputBraceLeft;
		if ( state == StateStandard ||
		     state == StateNumber ||
		     state == StateScalar ||
		     state == StateProtScalar ||
		     state == StateStartScalar ||
		     state == StateHash
                 )
		    paragData->parenList << Paren( Paren::Open, c, i );
		break;
	    case '}':
	        input = InputBraceRight;
		if ( state == StateStandard ||
		     state == StateNumber ||
		     state == StateScalar ||
		     state == StateProtScalar ||
		     state == StateStartScalar ||
		     state == StateHash
                  )
		    paragData->parenList << Paren( Paren::Closed, c, i );
		break;
	    case '/':
		input = InputSlash;
		break;
	    case '@':
		input = InputArobas;
		break;
	    case '%':
		input = InputPercent;
		break;
	    case '(': case '[':
		input = InputParen;
		if ( state == StateStandard ||
		     state == StateNumber ||
		     state == StateScalar )
		    paragData->parenList << Paren( Paren::Open, c, i );
		break;
	    case ')': case ']':
		input = InputParen;
		if ( state == StateStandard ||
		     state == StateNumber ||
		     state == StateScalar )
		    paragData->parenList << Paren( Paren::Closed, c, i );
		break;
	    case '#':
		input = InputHash;
		break;
	    case '"':
		input = InputQuotation;
		break;
	    case '\'':
		input = InputApostrophe;
		break;
	    case ' ':
		input = InputSpace;
		break;
	    case '1': case '2': case '3': case '4': case '5':
	    case '6': case '7': case '8': case '9': case '0':
		if ( alphabet.bsearch( lastChar ) != -1 &&
		     ( mathChars.find( lastChar ) == -1 || numbers.bsearch( string->at( i - 1 )->c ) == -1 ) ) {
		    input = InputAlpha;
		} else {
		    if ( input == InputAlpha && numbers.bsearch( lastChar ) != -1 )
			input = InputAlpha;
		    else
			input = InputNumber;
		}
		break;
	    case ':': {
		input = InputAlpha;
		QChar nextChar = ' ';
		if ( i < string->length() - 1 )
		    nextChar = string->at( i + 1 )->c;
		if ( state == StateStandard && !questionMark && lastChar != ':' && lastChar != ' ' && nextChar != ':') {
                    int j;
                    for (j=i-1; j>=0; --j){
                        if (string->at( j )->c.isSpace()) {
                            while(j && string->at( j )->c.isSpace())
                               --j;
                            if(j) goto BAILOUT;
                        } else if (alphabet.bsearch( string->at(j)->c ) == -1 && string->at(j)->c != '_')
                           goto BAILOUT;
                    }
		    for (++j; j < i; ++j ) {
			if ( string->at( j )->format() == formatStandard )
			    string->setFormat( j, 1, formatLabel, FALSE );
		    }
                    BAILOUT:;
		}
	    } break;
	    default: {
		if ( !questionMark && c == '?' )
		    questionMark = TRUE;
		if ( c.isLetter() || c == '_' )
		    input = InputAlpha;
		else
		    input = InputSep;
	    } break;
	    }
	}

	lastWasBackSlash = !lastWasBackSlash && c == '\\';

	if ( input == InputAlpha && !lastWasBackSlash )
	    buffer += c;
	int popAfter = 0;
	for ( ;; )
	{
	    int oldstate  = state;
	    state = table[ state ][ input ];
	    //qDebug( "newstate -> %d input->%d c->%c", state, input, c.latin1() );
	    if( state & Pop )
	    {
	        int n = state & PopMask;
		//qDebug( "npop: %d sp: %d",n, sp );
		if( state & After )
		{
		    popAfter = n;
		    state = oldstate;
		    break;
		}

		if(sp>=n)
		  sp -= n;
		else
		  sp = 0;
		state = stack[sp];
	    }
	    else if( state != stack[sp] )
	    {
		//bool stop = !(state & Again);
		state = state & PopMask;
		stack[++sp] = state;
		//qDebug( "sp: %d - st: %d", sp, state );
		if( sp == (int)stack.count()-1 )
		    sp = 0;
		//if( stop )
		    break;
	    }
	    else
		break;
	}

	switch ( state ) {
	case StateStandard: {
	    // future regexpr extension : use (?:((?:^|[;}])\s*|=~\s*)(m|s|tr))|[^\w\d_:@$%\\]q[qwxr]?(\W)
	    int len = buffer.length();
	    string->setFormat( i, 1, formatStandard, FALSE );
	    if ( makeLastStandard )
		string->setFormat( i - 1, 1, formatStandard, FALSE );
	    makeLastStandard = FALSE;
	    if ( len > 0 ) {
		if( input == InputAlpha && len < 3 && regExpMatcher->exactMatch( buffer ) ) {
		    state = StateMayBeRegExp;
		    stack[++sp] = state;
		    if( sp == (int)stack.count()-1 )
		        sp = 0;
		} else if ( input != InputAlpha ) {
		    if ( buffer.startsWith("Qt::") || buffer.startsWith("KDE::") ) {
		        string->setFormat( i - len, len, formatType, FALSE );
		    } else {
		        QMap<int, QMap<QString, int > >::Iterator it = wordMap->find( len );
			if ( it != wordMap->end() ) {
			    QMap<QString, int >::Iterator it2 = ( *it ).find( buffer );
			    if ( it2 != ( *it ).end() )
				string->setFormat( i - len, len, format( ( *it2 ) ), FALSE );
			}
		    }
		    buffer = QString::null;
		}
	    }
            if( input == InputSlash ) {
		// either a division sign or a RegExp matcher
		// it's almost impossible to tell for certain, so we must resort to some strong heuristics
		//int pos = string->string()->toString().findRev( *slashREMatcher, i+1 );
		//if( pos != -1 && i == (pos -1 + slashREMatcher->matchedLength()) ) {
		int j = checkSimpleRegExp( string, i );
		if( j )
		{
		    state = StateRegExp;
		    paragData->endChar= '/';
		    stack[++sp] = state;
		    if( sp == (int)stack.count()-1 )
			sp = 0;
		}
	    }

	} break;
	case StateScalar:
	case StateProtScalar:
	case StateStartScalar:
	    if ( makeLastStandard )
		string->setFormat( i - 1, 1, formatStandard, FALSE );
	    makeLastStandard = FALSE;
            string->setFormat( i, 1, formatScalar, FALSE );
	    buffer = QString::null;
	    break;
	case StateString:
	    if ( makeLastStandard )
		string->setFormat( i - 1, 1, formatStandard, FALSE );
	    makeLastStandard = FALSE;
	    if ( input == InputQuotation )
		string->setFormat( i, 1, formatStandard, FALSE );
	    else
		string->setFormat( i, 1, formatString, FALSE );
	    buffer = QString::null;
	    break;
	case StateString2:
	    if ( makeLastStandard )
		string->setFormat( i - 1, 1, formatStandard, FALSE );
	    makeLastStandard = FALSE;
	    if ( input == InputApostrophe )
		string->setFormat( i, 1, formatStandard, FALSE );
	    else
		string->setFormat( i, 1, formatString, FALSE );
	    buffer = QString::null;
	    break;
	case StateHash:
	    if ( makeLastStandard )
		string->setFormat( i - 1, 1, formatStandard, FALSE );
	    makeLastStandard = FALSE;
	    if ( input == InputBraceLeft || input == InputBraceRight )
		string->setFormat( i, 1, formatStandard, FALSE );
	    else
		string->setFormat( i, 1, formatHash, FALSE );
	    buffer = QString::null;
	    break;
	case StateNumber:
	    if ( makeLastStandard )
		string->setFormat( i - 1, 1, formatStandard, FALSE );
	    makeLastStandard = FALSE;
	    string->setFormat( i, 1, formatNumber, FALSE );
	    buffer = QString::null;
	    break;
	case StateComment:
	    if ( makeLastStandard )
		string->setFormat( i - 1, 1, formatStandard, FALSE );
	    makeLastStandard = FALSE;
	    string->setFormat( i, 1, formatComment, FALSE );
	    buffer = QString::null;
	    break;
	case StateMayBeRegExp:
	    string->setFormat( i, 1, formatStandard, FALSE );
	    paragData->endChar= balancedCharFor( c.latin1() );
	    state = stack[sp] = StateRegExp;
	    if( regExpMatcher->cap() == "s" || regExpMatcher->cap() == "tr" )
	    {
		stack[++sp] = StateRegExp;
		if( sp == (int)stack.count()-1 )
		    sp = 0;
	    }
	    buffer = QString::null;
	    break;
	case StateRegExp:
	    //if( sp>0 && stack[sp-1] != StateMayBeRegExp )
	    if( !(lastChar == '\\' && input == InputSep) && c.latin1() == paragData->endChar ) {
	        string->setFormat( i, 1, formatStandard, FALSE );
	//	paragData->endChar = 0;
		popAfter = 1;
	    } else
		string->setFormat( i, 1, formatRegExp, FALSE );
	    buffer = QString::null;
	    break;
	}
	if ( !!popAfter )
	{
		//qDebug( "popAfter: %d sp: %d",popAfter, sp );
		if(sp >= popAfter)
		  sp -= popAfter;
		else
		  sp = 0;
		state = stack[sp];
	}
	lastChar = c;
	i++;
	if ( i >= string->length() )
	    break;
    }

    alphabet.resetRawData( alpha, 52 );
    numbers.resetRawData( num, 10 );

    string->setExtraData( paragData );

    int oldEndState = string->endState();
//    if ( state == StateString ) {
//	string->setEndState( StateString );
//    } else if ( state == StateString2 ) {
//	string->setEndState( StateString2 );
//    } else if ( state == StateHash ) {
//	string->setEndState( StateHash );
    if ( state == StateComment && sp) {
	string->setEndState( stack[--sp] );
    } else {
//	string->setEndState( StateStandard );
//	sp = 0;
	string->setEndState( state );
    }

    string->setFirstPreProcess( FALSE );

    QTextParagraph *p = string->next();
    if ( (!!oldEndState || !!string->endState()) && oldEndState != string->endState() &&
	 invalidate && p && !p->firstPreProcess() && p->endState() != -1 ) {
	while ( p ) {
	    if ( p->endState() == -1 )
		return;
	    p->setEndState( -1 );
	    p = p->next();
	}
    }
}

QTextFormat *SyntaxHighlighter_Perl::format( int id )
{
    if ( lastFormatId == id  && lastFormat )
	return lastFormat;

    QTextFormat *f = formats[ id ];
    lastFormat = f ? f : formats[ 0 ];
    lastFormatId = id;
    return lastFormat;
}

void SyntaxHighlighter_Perl::addFormat( int id, QTextFormat *f )
{
    formats.insert( id, f );
}

void SyntaxHighlighter_Perl::removeFormat( int id )
{
    formats.remove( id );
}

